/****************************************************************************
**
** Copyright (C) 2005-2006 Trolltech AS. All rights reserved.
**
** This file is part of the example classes of the Qt Toolkit.
**
** This file may be used under the terms of the GNU General Public
** License version 2.0 as published by the Free Software Foundation
** and appearing in the file LICENSE.GPL included in the packaging of
** this file.  Please review the following information to ensure GNU
** General Public Licensing requirements will be met:
** http://www.trolltech.com/products/qt/opensource.html
**
** If you are unsure which license is appropriate for your use, please
** review the following information:
** http://www.trolltech.com/products/qt/licensing.html or contact the
** sales department at sales@trolltech.com.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
****************************************************************************/

#include <QtGui>
#include <QMenu>
#include <QSplashScreen>
#include <QProcess>
#include <QPixmap>
#include <QFile>
#include "mainwindow.h"
#include "browser.h"
#include "browsertreedbusextension.h"
#include "configurationfilewatcher.h"
#include "BrowserAdaptor.h"
#include "treestatemanager.h"
#include <TAppInfoList> /* used to retrieve the list of running applications when tree is reloaded */
#include <QDesktopServices>
#include <macros.h>
#include <QMessageBox>
#include <QMenuBar>
#include <QStatusBar>
#include <QFileDialog>

MainWindow::MainWindow(QWidget* parent_w)
{
    trayIcon = nullptr;
    QString splashFile = "elettra.png";
    int splashIdx = qApp->arguments().indexOf("--splash") + 1;
    if(splashIdx > 1)
        splashFile = qApp->arguments().at(splashIdx);

    QSplashScreen *splash = new QSplashScreen(QPixmap(splashFile));
    splash->showMessage("Loading Browser...", Qt::AlignHCenter);
    splash->show();
    resize(280, 580);

    browserTree = new BrowserTree(this);
    browserTree->setObjectName("BrowserTree");
    setCentralWidget(browserTree);
    setWindowIcon(QIcon(browserTree->SystrayIconFile()) );

    /* BrowserTree helper: manages DBus messages and modifies the tree */
    BrowserTreeDBusExtension *treeDBusExtension = new BrowserTreeDBusExtension(browserTree);
    /* session bus monitor, used to update the state of each application on the fly */
    connect(browserTree, SIGNAL(dbusRaise(const QString& )), treeDBusExtension, SLOT(dbusRaiseCumbia(const QString&)));
    connect(browserTree, SIGNAL(dbusStop(const QString& )), treeDBusExtension, SLOT(dbusStopQTangoCumbia(const QString&)));
    /* Get running applications from the DBus session and update the tree by means of the treeDBusExtension */
    QList<AppInfo> runningAppList = treeDBusExtension->refreshList();
    treeDBusExtension->setApplicationsRunning(runningAppList);
    /* end dbus initialization */

    createMenus();

    statusBar()->showMessage(tr("Ready"));

    if(browserTree->IsEditMode() )
        setWindowTitle("THE BROWSER IS IN EDIT MODE");
    else
        setWindowTitle(browserTree->WindowTitle() );
    
    /* SysTray stuff */
    CreateSysTrayIconActions();
   // CreateTrayIcon();
  //  connect(trayIcon, SIGNAL(activated(QSystemTrayIcon::ActivationReason)),
   //         this, SLOT(iconActivated(QSystemTrayIcon::ActivationReason)));
    // trayIcon->show();

//    if(browserTree->UniqueEnabled() )
//    {
//        trayIcon->setToolTip("Browser (with X11 unique application check enabled):"
//                             "\nnavigate inside the device tree and start the applications!");
//    }
//    else
//        trayIcon->setToolTip("Elettra browser: navigate inside the\n"
//                             "device tree and start the applications!");

    never_told_about_systray = true;

    connect(qApp, SIGNAL(Rise() ), this, SLOT(raise() ) );
    connect(qApp, SIGNAL(Rise() ), this, SLOT(show() ) );

    setMinimumSize(QSize(200, 300) );

    ConfigurationFileWatcher *xmlFileWatcher = new ConfigurationFileWatcher(this, browserTree->Filename());
    connect(xmlFileWatcher, SIGNAL(fileChanged()), this, SLOT(ReloadBrowser()));

    splash->finish(this);
    delete splash;
}

MainWindow::~MainWindow()
{
    if(trayIcon)
        delete trayIcon;
    delete browserTree;
}

void MainWindow::open()
{
    QString fileName =
            QFileDialog::getOpenFileName(this, tr("Open Bookmark File"),
                                         QDir::currentPath(),
                                         tr("XBEL Files (*.xbel *.xml)"));
    if (fileName.isEmpty())
        return;

    QFile file(fileName);
    if (!file.open(QFile::ReadOnly | QFile::Text)) {
        QMessageBox::warning(this, tr("SAX Bookmarks"),
                             tr("Cannot read file %1:\n%2.")
                             .arg(fileName)
                             .arg(file.errorString()));
        return;
    }

    if (browserTree->read(&file))
        statusBar()->showMessage(tr("File loaded"), 2000);
}

void MainWindow::saveAs()
{
    QString fileName =
            QFileDialog::getSaveFileName(this, tr("Save Bookmark File"),
                                         QDir::currentPath(),
                                         tr("XBEL Files (*.xbel *.xml)"));
    if (fileName.isEmpty())
        return;

    QFile file(fileName);
    if (!file.open(QFile::WriteOnly | QFile::Text)) {
        QMessageBox::warning(this, tr("SAX Bookmarks"),
                             tr("Cannot write file %1:\n%2.")
                             .arg(fileName)
                             .arg(file.errorString()));
        return;
    }

    if (browserTree->write(&file))
        statusBar()->showMessage(tr("File saved"), 2000);
}

void MainWindow::about()
{
    QMessageBox::about(this, tr("About Elettra Browser"),
                       QString("The <b>Browser</b> is an Elettra/Fermi resource browser "
                               "written by Giacomo Strangolino and based on the QT4 libraries.<br/>"
                               "It is based on an XML configuration file describing the elettra devices tree.<br/>"
                               "It supports the systray icon which allows to keep an eye on the browser<br/>"
                               "behaviour and maximize/minimize it without exiting from the application.<br/>"
                               "DBus communication guarantees that only one browser is running in the same machine.<br/>"
                               "<br/>Version %1<br/>"
                               "DBus session bus allows the communication between the browser and the QTango applications."
                               "Write to <a href=\"mailto:giacomo.strangolino@elettra.trieste.it\">giacomo.strangolino@elettra.trieste.it </a> for information or help.<br/>June - November 2006.")
                       .arg(qApp->applicationVersion()));
}

QStringList MainWindow::runningApplications() const
{
    return browserTree->runningApplications() ;
}

void MainWindow::createMenus()
{
    QList <QPair <bool, QString> > execs;
    QIcon icon;
    //    openAct = new QAction(tr("&Open..."), this);
    //    openAct->setShortcut(tr("Ctrl+O"));
    //    connect(openAct, SIGNAL(triggered()), this, SLOT(open()));

    fileMenu = menuBar()->addMenu("&File");
    fileMenu->setObjectName("File Menu");
    menuBar()->setObjectName("Menu Bar");

    
    if(browserTree->IsEditMode() )
    {
        saveAsAct = new QAction(tr("&Save As..."), this);
        saveAsAct->setShortcut(tr("Ctrl+S"));
        connect(saveAsAct, SIGNAL(triggered()), this, SLOT(saveAs()));
        fileMenu->addAction(saveAsAct );
    }
    
    restartAction = new QAction("Reload tree", this);
    restartAction->setIcon(QIcon(":icons/reload.png"));
    restartAction->setObjectName("Restart Action");
    restartAction->setShortcut(tr("Ctrl+R"));
    connect(restartAction, SIGNAL(triggered() ), this, SLOT(ReloadBrowser() ) );
    fileMenu->addAction(restartAction);

    fileMenu->addSeparator();

    exitAct = new QAction(tr("E&xit"), this);
    exitAct->setShortcut(tr("Ctrl+Q"));
    exitAct->setIcon(QIcon(":icons/exit.png") );
    connect(exitAct, SIGNAL(triggered()), qApp, SLOT(quit()));
    fileMenu->addAction(exitAct);
    

    aboutAct = new QAction(tr("&About"), this);
    connect(aboutAct, SIGNAL(triggered()), this, SLOT(about()));

    aboutQtAct = new QAction(tr("About &Qt"), this);
    connect(aboutQtAct, SIGNAL(triggered()), qApp, SLOT(aboutQt()));
    
    /* View menu */
    viewMenu = menuBar()->addMenu("&View");
    viewMenu->setObjectName("View Menu");
    /* View actions */
    viewExecHeader = new QAction("&View execute command", this);
    viewExecHeader->setCheckable(true);
    viewExecHeader->setChecked(false);
    connect(viewExecHeader , SIGNAL(toggled(bool) ), browserTree, SLOT( view_exec_command(bool) ) );
    viewMenu->addAction(viewExecHeader);
    viewStateHeader = new QAction("View Application &States", this);
    viewStateHeader->setCheckable(true);
    viewStateHeader->setChecked(false);
    connect(viewStateHeader, SIGNAL(triggered(bool) ), browserTree, SLOT(view_appli_states(bool) ) );
    viewMenu->addAction(viewStateHeader);
    viewDescriptionHeader = new QAction("View &Description", this);
    viewDescriptionHeader->setCheckable(true);
    viewDescriptionHeader->setChecked(false);
    connect(viewDescriptionHeader, SIGNAL(triggered(bool) ), browserTree, SLOT(view_description(bool) ) );
    viewMenu->addAction(viewDescriptionHeader);


    unlockHeadersAction = new QAction("Unlock Headers", this);
    unlockHeadersAction->setCheckable(true);
    connect(unlockHeadersAction, SIGNAL(triggered(bool)), browserTree, SLOT(unlockHeaders(bool)));
    viewDBusServiceAction =  new QAction("Dbus Service", this);
    viewDBusServiceAction->setCheckable(true);
    connect(viewDBusServiceAction, SIGNAL(triggered(bool)), browserTree, SLOT(viewDBusService(bool)));

    viewMenu->addAction(viewDBusServiceAction);

    viewMenu->addSeparator();

    viewMenu->addAction(unlockHeadersAction);

    viewHtml = new QAction("View &html rendering (help)", this);
    viewHtml->setIcon(QIcon(":icons/browser.png") );
    connect(viewHtml, SIGNAL(triggered() ), this, SLOT(view_html() ) );
    viewMenu->addAction(viewHtml);

    toolsMenu = menuBar()->addMenu("&Tools");
    toolsMenu->setObjectName("Tools Menu");

    findAction = new QAction("&Find an entry...", this);
    QKeySequence seq("Ctrl+F");
    findAction->setShortcut(seq);
    findAction->setIcon(QIcon(":icons/find.png") );
    connect(findAction, SIGNAL(triggered() ), browserTree, SLOT(ShowFindDialog() ) );
    toolsMenu->addAction(findAction);

    /* Help menu */
    helpMenu = menuBar()->addMenu(tr("&Help"));
    helpMenu->setObjectName("Help Menu");
    
    helpIconAct = helpMenu->addMenu(browserTree->IconMeaningsMenu() );
    helpIconAct->setIcon(QIcon(":icons/help.png") );
    helpIconAct->setText("Icons Legend");
    
    QAction* viewHtmlHelp = new QAction("HTML Browser Tree Help", this);
    helpMenu->addAction(viewHtmlHelp);
    connect(viewHtmlHelp, SIGNAL(triggered() ), this, SLOT(view_html() ) );

    QMenu *problems = helpMenu->addMenu("Problems...");
    QAction *killact = problems->addAction("How to stop a panel not responding...");
    connect(killact, SIGNAL(triggered() ), this, SLOT(killmenu() ) );
    
    aboutAct = new QAction(tr("&About"), this);
    connect(aboutAct, SIGNAL(triggered()), this, SLOT(about()));
    helpMenu->addAction(aboutAct);
    
    aboutQtAct = new QAction(tr("About &Qt"), this);
    connect(aboutQtAct, SIGNAL(triggered()), qApp, SLOT(aboutQt()));
    helpMenu->addAction(aboutQtAct);
}

void MainWindow::view_html()
{
    QString service = "file://" + browserTree->Filename();
    printf("launching service \"%s\"\n", qstoc(service));
    QDesktopServices::openUrl(QUrl(service));
}

void MainWindow::killmenu()
{
    QMessageBox::information(this, "NEW! You do not need to use XKILL anymore!",
                             "Thanks to this new version of the browser you won't need to XKILL\n"
                             "anymore a window that is not responding!\n"
                             "Just locate the freezed application inside the browser tree and:\n"
                             "1. Right click on the item;\n"
                             "2. Move your mouse over the label \"Stop processes\": a new menu will appear.\n"
                             "3a. Try the first option: \"Safely close the application\": this should close\n"
                             "    the freezed panel. If it does not work, try with 3b;\n"
                             "3b. Try INTERRUPTing it: an INTERRUPT signal is sent to the application: it should\n"
                             "    intercept it and quit: click on the \"INTERRUPT\" option;\n"
                             "    If the panel is still locked, you've got the last chance to kill it, with the\n"
                             "    KILL signal: you should avoid it possibly, anyway:\n"
                             "3c. Click on \"BRUTALLY KILL\" option to force the application to quit.\n"
                             "    This option should always work. If it does not, contact the administrator and\n"
                             "    DO NOT use xkill to terminate the window which behaves badly.\n"
                             "Actually `xkill' does not always stop the application, it just kills the connection\n"
                             "between the window and the graphical server (X server).\n");
}

void MainWindow::closeEvent(QCloseEvent* e)
{
    int ret = QMessageBox::question(this, "Quit the browser application?",
                             "Are you sure you want to quit the application?");
    if(ret == QMessageBox::Yes)
        QMainWindow::closeEvent(e);
    else
        e->ignore();
}

void MainWindow::ReloadBrowser()
{
    /* First of all try to see if there are problems with the xml */
    QFile xmlfile(browserTree->Filename());
    QString errorStr;
    int errorLine;
    int errorColumn;
    QDomDocument domdoc_test;

    if (!domdoc_test.setContent(&xmlfile, true, &errorStr, &errorLine,
                                &errorColumn))
    {
        QMessageBox::information(this, tr("Browser"),
                                 tr("Parse error at line %1, column %2: (while reloading the tree)\n%3")
                                 .arg(errorLine)
                                 .arg(errorColumn)
                                 .arg(errorStr));
        if(trayIcon)
            trayIcon->showMessage(QString("%1").arg(browserTree->WindowTitle()),
                              "The browser tree could not be reloaded\nbecause an error in the xml"
                              " configuration\nfile happened.");

        return;
    }
    TreeStateManager stateManager;
    stateManager.saveState(browserTree);

    delete browserTree->findChild<BrowserTreeDBusExtension *>();
    delete browserTree;
    browserTree = new BrowserTree(parentWidget() );
    setCentralWidget(browserTree);
    setWindowIcon(QIcon(browserTree->SystrayIconFile()) );

    statusBar()->showMessage(tr("Browser reloaded"));

    if(browserTree->IsEditMode() )
        setWindowTitle("THE BROWSER IS IN EDIT MODE");
    else
        setWindowTitle(browserTree->WindowTitle() );

    connect(viewExecHeader , SIGNAL(toggled(bool) ), browserTree, SLOT( view_exec_command(bool) ) );
    connect(viewStateHeader, SIGNAL(toggled(bool) ), browserTree, SLOT(view_appli_states(bool) ) );
    connect(viewDescriptionHeader, SIGNAL(toggled(bool)), browserTree, SLOT(view_description(bool) ) );
    connect(findAction, SIGNAL(triggered() ), browserTree, SLOT(ShowFindDialog() ) );
    browserTree->view_exec_command(viewExecHeader->isChecked());
    browserTree->view_appli_states(viewStateHeader->isChecked());
    browserTree->view_description(viewDescriptionHeader->isChecked());

    if(!browserTree->UniqueEnabled() && trayIcon )
        trayIcon->showMessage(QString("%1: warning").arg(browserTree->WindowTitle() ),
                              "The tree has been automatically reloaded.\nThe unique application check is disabled");
    else if(trayIcon)
        trayIcon->showMessage(QString("%1").arg(browserTree->WindowTitle()),
                              "The browser tree has been automatically reloaded.");
    stateManager.restoreState(browserTree);
    /* restore state for running applications */
    /* BrowserTree helper: manages DBus messages and modifies the tree. Must recreate a new instance
         * since the old one has been deleted together with old browserTree (TreeDBusExtension is  child of browserTree
         */
    BrowserTreeDBusExtension *treeDBusExtension = new BrowserTreeDBusExtension(browserTree);
    QList<AppInfo> runningAppList = treeDBusExtension->refreshList();
    treeDBusExtension->setApplicationsRunning(runningAppList);
    connect(browserTree, SIGNAL(dbusRaise(const QString& )), treeDBusExtension, SLOT(dbusRaiseCumbia(const QString&)));
    connect(browserTree, SIGNAL(dbusStop(const QString& )), treeDBusExtension, SLOT(dbusStopQTangoCumbia(const QString&)));
}

void MainWindow::CreateSysTrayIconActions()
{
    minimizeAction = new QAction(tr("Mi&nimize"), this);
    connect(minimizeAction, SIGNAL(triggered()), this, SLOT(hide()));

    maximizeAction = new QAction(tr("Ma&ximize"), this);
    connect(maximizeAction, SIGNAL(triggered()), this, SLOT(showMaximized()));

    restoreAction = new QAction(tr("&Restore"), this);
    connect(restoreAction, SIGNAL(triggered()), this, SLOT(showNormal()));

    searchSystrayAction = new QAction(tr("&Find..."), this);
    connect(searchSystrayAction, SIGNAL(triggered()), this, SLOT(find_from_systray()));

    quitAction = new QAction(tr("&Quit"), this);
    connect(quitAction, SIGNAL(triggered()), qApp, SLOT(quit()) );
}

void MainWindow::CreateTrayIcon()
{
    trayIconMenu = new QMenu(this);
    trayIconMenu->setObjectName("trayIconMenu");
    trayIconMenu->addSeparator();
    minimizeAction->setIcon(QIcon(":icons/go-down.png"));
    maximizeAction->setIcon(QIcon(":icons/view_restore.png"));
    restoreAction->setIcon(QIcon(":icons/view-list-tree.png"));
    quitAction->setIcon(QIcon(":icons/exit.png"));
    trayIconMenu->addAction(minimizeAction);
    trayIconMenu->addAction(maximizeAction);
    trayIconMenu->addAction(restoreAction);
    trayIconMenu->addSeparator();
    searchSystrayAction->setIcon(QIcon(":icons/find.png" ) );
    trayIconMenu->addAction(searchSystrayAction);
    trayIconMenu->addSeparator();
    trayIconMenu->addAction(quitAction);

    if(!QSystemTrayIcon::isSystemTrayAvailable()) {
        printf("\e[1;31m!! \e[0m the system tray is not available on this platform\n\n");
    }

    trayIcon = new QSystemTrayIcon(this);
    trayIcon->setContextMenu(trayIconMenu);
    QIcon icon(browserTree->SystrayIconFile() );
    trayIcon->setIcon(icon);
}

void MainWindow::find_from_systray()
{
    int ret = 0;
    if( (ret = browserTree->ShowFindDialog() )  > 0)
    {
        show();
        trayIcon->showMessage("Search results", QString("Found %1 elements in the tree").arg(ret) );
    }
    else
        /* otherwise do not show() the window */
        trayIcon->showMessage("Search results", "No elements found matching your search!");
}

void MainWindow::raiseFromDBusMessage()
{
    if(isVisible())
        hide();
    /* then show, in any case */
    showNormal();
}

void MainWindow::iconActivated(QSystemTrayIcon::ActivationReason reason)
{
    //	qDebug() << "stato della finestra: " << windowState();
    switch (reason) {
    case QSystemTrayIcon::Trigger:
    case QSystemTrayIcon::DoubleClick:
        never_told_about_systray = false;
        if(isHidden() )
            ; //	qDebug() << "e` hidden";
        else
        {
            //		trayIcon->showMessage("Browser already visible",
            //			"The browser should be already visible\n"
            //			"somewhere, anyway, let's try to show it\n"
            //			"above other windows...",
            //			QSystemTrayIcon::Information, 3000);
            //		qDebug() << "E` visible";
            hide();
        }

        showNormal();
        setWindowState(Qt::WindowActive);
        //	else
        //		hide();
        break;
    default:
        //	qDebug() << "altra reason, no Trigger ne` DoubleClick";
        ;
    }
}
